﻿using System;
using Ozeki.Media;
using Ozeki.MediaGateway;
using Ozeki.VoIP;

namespace SIPGateway
{
    class SIPClient
    {
        IClient client;

        ISoftPhone softPhone;
        IPhoneLine phoneLine;
        IPhoneCall call;

        IStreamService streamService;
        string publishStreamName;
        string playStreamName;

        MediaConnector mediaConnector;
        PhoneCallAudioSender phoneCallAudioSender;
        PhoneCallAudioReceiver phoneCallAudioReceiver;

        MediaGatewayAudioReceiver mediaGatewayAudioReceiver;
        MediaGatewayAudioSender mediaGatewayAudioSender;


        /// <summary>
        /// Creates a new instance of SIPClient.
        /// </summary>
        /// <param name="client">The connected client.</param>
        /// <param name="streamService"> Stream service what transmits media data between two client.</param>
        /// <param name="softPhone"> Represents a SIP softphone</param>
        /// <param name="account">SIP account for PBX registration.</param>
        public SIPClient(IClient client, IStreamService streamService, ISoftPhone softPhone, SIPAccount account)
        {
            mediaConnector = new MediaConnector();
            phoneCallAudioSender = new PhoneCallAudioSender();
            phoneCallAudioReceiver = new PhoneCallAudioReceiver();

            this.streamService = streamService;
            this.softPhone = softPhone;
            this.client = client;

            Account = account;
        }

        /// <summary>
        /// SIP account for able to register a SIP PBX.
        /// </summary>
        public SIPAccount Account { get; protected set; }

        /// <summary>
        /// Indicates phoneline registration changes.
        /// </summary>
        void PhoneLine_RegistrationStateChanged(object sender, RegistrationStateChangedArgs e)
        {
            if (e.State == RegState.RegistrationSucceeded)
                client.InvokeMethod("OnRegisteredUserNameReceived", Account.DisplayName);

            OnPhoneLineStateChanged(e.State);
        }

        /// <summary>
        /// Indicates received incoming call.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        void SoftPhone_IncomingCall(object sender, VoIPEventArgs<IPhoneCall> e)
        {
            if (e.Item.PhoneLine == phoneLine) 
            {
                call = e.Item;
                call.CallStateChanged += Call_StateChanged;
                call.DtmfReceived += call_DtmfReceived;
                OnIncomingCall(call.DialInfo.CallerID);
            }
        }


        /// <summary>
        /// Call the given dial number.
        /// </summary>
        /// <param name="dialNumber">dial number</param>
        public virtual void Call(string dialNumber)
        {
            if (phoneLine == null)
            {
                System.Diagnostics.Debug.Fail("PhoneLine null");
                return;
            }

            if ((phoneLine.RegState == RegState.RegistrationSucceeded) && call == null)
            {
                call = softPhone.CreateCallObject(phoneLine, dialNumber);
                call.CallStateChanged += Call_StateChanged;
                call.DtmfReceived += call_DtmfReceived;
                call.Start();
            }
        }

        /// <summary>
        /// Indicates call state changes.
        /// </summary>
        protected virtual void Call_StateChanged(object sender, CallStateChangedArgs e)
        {
            if (e.State == CallState.Answered)
            {
                phoneCallAudioSender.AttachToCall(call);
                phoneCallAudioReceiver.AttachToCall(call);

                publishStreamName = Guid.NewGuid().ToString();
                playStreamName = Guid.NewGuid().ToString();

                var mediaStream = streamService.CreateStream(playStreamName);
                mediaGatewayAudioReceiver = new MediaGatewayAudioReceiver(mediaStream);

                mediaConnector.Connect(phoneCallAudioReceiver, mediaGatewayAudioReceiver);

                OnInCall(publishStreamName, playStreamName);
            }

            if (e.State.IsCallEnded())
            {
                phoneCallAudioSender.Detach();
                phoneCallAudioReceiver.Detach();

                streamService.RemoveStream(mediaGatewayAudioReceiver.MediaStream);
                mediaConnector.Disconnect(phoneCallAudioReceiver, mediaGatewayAudioReceiver);
                mediaGatewayAudioReceiver.Dispose();

                mediaGatewayAudioReceiver = null;
                call = null;
            }

            OnCallStateChanged(e.State);
        }

        /// <summary>
        /// Indicates receiving DTMF signals.
        /// </summary>
        protected void call_DtmfReceived(object sender, VoIPEventArgs<DtmfInfo> e)
        {
            OnCallDTMFReceived(e.Item.Signal);
        }

        /// <summary>
        /// Hangup the call.
        /// </summary>
        public virtual void HangUp()
        {
            if (call != null)
            {
                call.HangUp();
                call = null;
            }
        }

        /// <summary>
        /// Accept the incoming call request.
        /// </summary>
        public virtual void Accept()
        {
            if (call != null)
                call.Answer();
        }

        /// <summary>
        /// Reject the incoming call request.
        /// </summary>
        public void Reject()
        {
            if (call != null)
            {
                call.Reject();
                call = null;
            }
        }

        /// <summary>
        /// Stats to send a DTMF sign.
        /// </summary>
        /// <param name="dtmf">DTMF sign</param>
        public void StartDtmf(DtmfNamedEvents dtmf)
        {
            if (call != null)
                call.StartDTMFSignal(dtmf);
        }

        /// <summary>
        /// Stops to send a DTMF sign.
        /// </summary>
        /// <param name="dtmf"></param>
        public void StopDtmf(DtmfNamedEvents dtmf)
        {
            if (call != null)
                call.StopDTMFSignal(dtmf);
        }

		/// <summary>
        /// Registers a SIP account for a SIP PBX.
        /// </summary>
        public virtual void Register()
        {
            phoneLine = softPhone.CreatePhoneLine(Account);
            phoneLine.RegistrationStateChanged += PhoneLine_RegistrationStateChanged;

            softPhone.IncomingCall += SoftPhone_IncomingCall;

            softPhone.RegisterPhoneLine(phoneLine);
        }

        /// <summary>
        /// Unregisters the registered phoneline from the SIP PBX. 
        /// </summary>
        public virtual void Unregister()
        {
            if (mediaGatewayAudioReceiver != null)
            {
                streamService.RemoveStream(mediaGatewayAudioReceiver.MediaStream);
                mediaGatewayAudioReceiver = null;
            }

            HangUp();

            softPhone.IncomingCall -= SoftPhone_IncomingCall;

            if (phoneLine != null)
            {
                phoneLine.RegistrationStateChanged -= PhoneLine_RegistrationStateChanged;
                softPhone.UnregisterPhoneLine(phoneLine);
            }
        }

        #region Client-side method invoke

        /// <summary>
        /// Invoke client's callback method. Notifies the client that call state has changed.
        /// </summary>
        void OnCallStateChanged(CallState callState)
        {
            client.InvokeMethod("OnCallStateChanged", callState);
        }

        /// <summary>
        /// Invoke client's callback method. Sends the publish and receive stream ID.
        /// </summary>
        void OnInCall(string publishName, string playName)
        {
            client.InvokeMethod("OnInCall", publishName, playName);
        }

        /// <summary>
        /// Invokes client's callback method. Notifies the client that it received a new call from another extension.
        /// </summary>
        /// <param name="phoneNumber">The request owner's phonenumber.</param>
        void OnIncomingCall(string phoneNumber)
        {
            client.InvokeMethod("OnIncomingCall", phoneNumber);
        }

        /// <summary>
        /// Invokes client's callback method. Sends the registration state change.
        /// </summary>
        /// <param name="regState"></param>
        void OnPhoneLineStateChanged(RegState regState)
        {
            client.InvokeMethod("OnPhoneLineStateChanged", regState);
        }

        /// <summary>
        /// Invokes client's callback method. Notifies the client that a DTMF signal received from the remote end.
        /// </summary>
        /// <param name="dtmfsignal"></param>
        void OnCallDTMFReceived(DtmfSignal dtmfsignal)
        {
            client.InvokeMethod("OnDTMFReceived",dtmfsignal.Signal);
        }

        #endregion

        public void StreamPublished(IMediaStream mediaStream)
        {
            mediaGatewayAudioSender = new MediaGatewayAudioSender(mediaStream);
            mediaConnector.Connect(mediaGatewayAudioSender, phoneCallAudioSender);
        }

        public void StreamClosed()
        {
            if (mediaGatewayAudioSender != null)
            {
                mediaConnector.Disconnect(mediaGatewayAudioSender, phoneCallAudioSender);
                mediaGatewayAudioSender.Dispose();
                mediaGatewayAudioSender = null;
            }
        }
    }
}
